@@ -9,13 +9,13 @@ module BackendList (allBackends) where
 
 -- When adding a new backend, import it here and add it to the list.
 import qualified Backend.WORM
-import qualified Backend.SHA1
+import qualified Backend.SHA
 import qualified Backend.URL
 import Types
 
-allBackends :: [Backend]
-allBackends = 
-	[ Backend.WORM.backend
-	, Backend.SHA1.backend
-	, Backend.URL.backend
+allBackends :: [Backend Annex]
+allBackends = concat 
+	[ Backend.WORM.backends
+	, Backend.SHA.backends
+	, Backend.URL.backends
 	]
