@@ -13,16 +13,20 @@ import Command
 import qualified Backend
 import LocationLog
 import Types
-import Core
+import Content
 import Messages
 import Utility
 
-seek :: [SubCmdSeek]
+command :: [Command]
+command = [repoCommand "drop" paramPath seek
+	"indicate content of files not currently wanted"]
+
+seek :: [CommandSeek]
 seek = [withAttrFilesInGit "annex.numcopies" start]
 
 {- Indicates a file's content is not wanted anymore, and should be removed
  - if it's safe to do so. -}
-start :: SubCmdStartAttrFile
+start :: CommandStartAttrFile
 start (file, attr) = isAnnexed file $ \(key, backend) -> do
 	inbackend <- Backend.hasKey key
 	if not inbackend
@@ -33,14 +37,14 @@ start (file, attr) = isAnnexed file $ \(key, backend) -> do
 	where
 		numcopies = readMaybe attr :: Maybe Int
 
-perform :: Key -> Backend -> Maybe Int -> SubCmdPerform
+perform :: Key -> Backend Annex -> Maybe Int -> CommandPerform
 perform key backend numcopies = do
 	success <- Backend.removeKey backend key numcopies
 	if success
 		then return $ Just $ cleanup key
 		else return Nothing
 
-cleanup :: Key -> SubCmdCleanup
+cleanup :: Key -> CommandCleanup
 cleanup key = do
 	inannex <- inAnnex key
 	when inannex $ removeAnnex key
