@@ -10,14 +10,18 @@ module Command.Get where
 import Command
 import qualified Backend
 import Types
-import Core
+import Content
 import Messages
 
-seek :: [SubCmdSeek]
+command :: [Command]
+command = [repoCommand "get" paramPath seek
+		"make content of annexed files available"]
+
+seek :: [CommandSeek]
 seek = [withFilesInGit start]
 
 {- Gets an annexed file from one of the backends. -}
-start :: SubCmdStartString
+start :: CommandStartString
 start file = isAnnexed file $ \(key, backend) -> do
 	inannex <- inAnnex key
 	if inannex
@@ -26,7 +30,7 @@ start file = isAnnexed file $ \(key, backend) -> do
 			showStart "get" file
 			return $ Just $ perform key backend
 
-perform :: Key -> Backend -> SubCmdPerform
+perform :: Key -> Backend Annex -> CommandPerform
 perform key backend = do
 	ok <- getViaTmp key (Backend.retrieveKeyFile backend key)
 	if ok
