@@ -8,8 +8,9 @@
 module Command.Init where
 
 import Control.Monad.State (liftIO)
-import Control.Monad (when)
+import Control.Monad (when, unless)
 import System.Directory
+import System.FilePath
 
 import Command
 import qualified Annex
@@ -19,34 +20,51 @@ import Version
 import Messages
 import Locations
 import Types
+import Utility
+	
+command :: [Command]
+command = [repoCommand "init" paramDesc seek
+		"initialize git-annex with repository description"]
 
-seek :: [SubCmdSeek]
+seek :: [CommandSeek]
 seek = [withString start]
 
 {- Stores description for the repository etc. -}
-start :: SubCmdStartString
+start :: CommandStartString
 start description = do
 	when (null description) $
 		error "please specify a description of this repository\n"
 	showStart "init" description
 	return $ Just $ perform description
 
-perform :: String -> SubCmdPerform
+perform :: String -> CommandPerform
 perform description = do
 	g <- Annex.gitRepo
 	u <- getUUID g
-	describeUUID u description
 	setVersion
-	liftIO $ gitAttributesWrite g
-	gitPreCommitHookWrite g
-	return $ Just cleanup
+	if Git.repoIsLocalBare g
+		then do
+			showLongNote $
+				"This is a bare repository, so its description cannot be committed.\n" ++
+				"To record the description, run this command in a clone of this repository:\n" ++
+				"   git annex describe " ++ show u ++ " " ++ show description ++ "\n\n"
+			return $ Just $ return True
+		else do
+			describeUUID u description
+			liftIO $ gitAttributesWrite g
+			gitPreCommitHookWrite g
+			return $ Just cleanup
 
-cleanup :: SubCmdCleanup
+cleanup :: CommandCleanup
 cleanup = do
 	g <- Annex.gitRepo
 	logfile <- uuidLog
-	liftIO $ Git.run g ["add", logfile]
-	liftIO $ Git.run g ["commit", "-m", "git annex init", logfile]
+	liftIO $ Git.run g "add" [File logfile]
+	liftIO $ Git.run g "commit" 
+		[ Params "-q -m"
+		, Param "git annex repository description"
+		, File logfile
+		]
 	return True
 
 {- configure git to use union merge driver on state files, if it is not
@@ -56,22 +74,30 @@ gitAttributesWrite repo = do
 	exists <- doesFileExist attributes
 	if not exists
 		then do
-			writeFile attributes $ attrLine ++ "\n"
+			safeWriteFile attributes $ unlines attrLines
 			commit
 		else do
 			content <- readFile attributes
-			when (all (/= attrLine) (lines content)) $ do
-				appendFile attributes $ attrLine ++ "\n"
+			let present = lines content
+			let missing = filter (\l -> not $ l `elem` present) attrLines
+			unless (null missing) $ do
+				appendFile attributes $ unlines missing
 				commit
 	where
 		attributes = Git.attributes repo
 		commit = do
-			Git.run repo ["add", attributes]
-			Git.run repo ["commit", "-m", "git-annex setup", 
-					attributes]
+			Git.run repo "add" [Param attributes]
+			Git.run repo "commit"
+				[ Params "-q -m"
+				, Param "git-annex setup"
+				, Param attributes
+				]
 
-attrLine :: String		
-attrLine = stateLoc ++ "*.log merge=union"
+attrLines :: [String]
+attrLines =
+	[ stateDir </> "*.log merge=union"
+	, stateDir </> "*/*/*.log merge=union"
+	]
 
 {- set up a git pre-commit hook, if one is not already present -}
 gitPreCommitHookWrite :: Git.Repo -> Annex ()
@@ -80,7 +106,7 @@ gitPreCommitHookWrite repo = do
 	if exists
 		then warning $ "pre-commit hook (" ++ hook ++ ") already exists, not configuring"
 		else liftIO $ do
-			writeFile hook preCommitScript
+			safeWriteFile hook preCommitScript
 			p <- getPermissions hook
 			setPermissions hook $ p {executable = True}
 	where
